{-# LANGUAGE RankNTypes #-}
module IGraph.Random
    ( Seed
    , Gen(..)
    , withSeed
    , withDefaultRng
    ) where

import           Control.Concurrent.MVar
import           Control.Exception.Safe
import           Foreign
import           IGraph.Internal
import           System.IO.Unsafe
import qualified Control.Monad as M

-- | Random number generator
newtype Gen s = Gen { _Gen :: Ptr RNG }

{-| Making RNG management in igraph thread safe

Unfortunately in igraph there is a global RNG which gets used by
/ALL/ the functions within igraph: changing it via igraphRngSetDefault
affects all the functions, which makes impossible to run concurrent
programs without bugs. This is documented in the igraph C Library:

 * <para>
 * Note that initializing a new random number generator is
 * independent of the generator that the igraph functions themselves
 * use. If you want to replace that, then please use \ref
 * igraph_rng_set_default().
 * </para>
 * </section>

Versions in the \"develop\" branch of igraph have modified
igraph_rng_set_default to return the pointer to the old RNG, but
version 0.11 is currently unreleased, so we have to use a C compat
shim to achieve the same.

However, that still doesn't fix the concurrency problem, so we use
a global lock (essentially, an MVar) to make sure that every call
to 'withSeed' will have exclusive access to the global lock.
-}

newtype GlobalRNGLock = GlobalRNGLock { _GlobalRNGLock :: MVar () }

globalRNGLock :: GlobalRNGLock
globalRNGLock = unsafePerformIO (GlobalRNGLock <$> newMVar ())
{-# NOINLINE globalRNGLock #-}

-- | Execute the inner action with exclusive access to the global RNG.
withGlobalRNGLock :: IO a -> IO a
withGlobalRNGLock act = withMVar (_GlobalRNGLock globalRNGLock) (\() -> act)
{-# INLINE withGlobalRNGLock #-}

-- | Creates a new 'Gen' with the input 'Seed', passing it
-- to the inner function 'fun'. The use of the existential 's'
-- ensures that we can't leak the generator outside the function.
withSeed :: Seed -> (forall s. Gen s -> IO a) -> IO a
withSeed seed fun = withGlobalRNGLock $ do
  allocaRng $ \rng -> do
    igraphRngSeed rng seed
    old <- igraphRngSetDefault rng
    (fun $ Gen rng) `finally` (M.void $ igraphRngSetDefault old)
{-# NOINLINE withSeed #-}

-- | Non-thread-safe function that /globally mutates/ the default
-- RNG with the provided seed. Use with extreme care.
withDefaultRng :: Seed -> (forall s. Gen s -> IO a) -> IO a
withDefaultRng seed fun = do
  rng <- igraphRngDefault
  igraphRngSeed rng seed
  (fun $ Gen rng)
{-# INLINE withDefaultRng #-}
